package com.sd.business.service.bom.impl;

import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import com.baomidou.dynamic.datasource.annotation.DSTransactional;
import com.baomidou.mybatisplus.core.toolkit.IdWorker;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.ruoyi.common.core.domain.BaseIdPo;
import com.sd.business.entity.bom.dto.BomDto;
import com.sd.business.entity.bom.dto.BomSelectDto;
import com.sd.business.entity.bom.po.Bom;
import com.sd.business.entity.bom.po.BomSpec;
import com.sd.business.entity.bom.vo.BomVo;
import com.sd.business.mapper.bom.BomMapper;
import com.sd.business.service.bom.BomClassifyService;
import com.sd.business.service.bom.BomService;
import com.sd.business.service.bom.BomSpecService;
import com.sd.framework.util.Assert;
import com.sd.framework.util.sql.Sql;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;


/**
 * <p>
 * bom 服务实现类
 * </p>
 *
 * @author
 * @since 2023-12-04
 */
@Service
public class BomServiceImpl extends ServiceImpl<BomMapper, Bom> implements BomService {

    @Autowired
    private BomClassifyService bomClassifyService;

    @Autowired
    private BomSpecService bomSpecService;

    @Override
    public Page<BomVo> getPage(BomSelectDto dto) {

        List<Long> queryBomIdList = null;
        Set<Long> queryBomClassifyIdSet = new HashSet<>();
        if (!StrUtil.isAllBlank(dto.getColour(), dto.getBomSpecCode(), dto.getBomSpecName())) {
            queryBomIdList = bomSpecService.listObject(BomSpec::getBomId, i -> i
                    .like(StrUtil.isNotBlank(dto.getColour()), BomSpec::getColour, dto.getColour())
                    .like(StrUtil.isNotBlank(dto.getBomSpecCode()), BomSpec::getCode, dto.getBomSpecCode())
                    .like(StrUtil.isNotBlank(dto.getBomSpecName()), BomSpec::getName, dto.getBomSpecName()));
            if (queryBomIdList.isEmpty()) {
                return new Page<>();
            }
        }
        if (ObjectUtil.isNotNull(dto.getBomClassifyId())) {
            queryBomClassifyIdSet.addAll(bomClassifyService.getChildrenIdList(dto.getBomClassifyId()));
        }
        if (ObjectUtil.isNotNull(dto.getBomClassifyIdList())) {
            queryBomClassifyIdSet.addAll(bomClassifyService.getChildrenIdList(dto.getBomClassifyIdList()));
        }

        Page<BomVo> page = Sql.create(BomVo.class)
                .selectAll(Bom.class)
                .from(Bom.class)
                .orderByDesc(Bom::getId)
                .like(Bom::getName, dto.getName())
                .like(Bom::getCode, dto.getCode())
                .eq(Bom::getSpecies, dto.getSpecies())
                .in(Bom::getId, queryBomIdList)
                .in(Bom::getBomClassifyId, queryBomClassifyIdSet)
                .page(dto);
        List<BomVo> records = page.getRecords();
        if (records.isEmpty()) {
            return page;
        }

        List<Long> bomIdList = records.stream().map(BaseIdPo::getId).collect(Collectors.toList());
        Map<Long, List<BomSpec>> bomSpecIdMap = bomSpecService.mapKGroup(BomSpec::getBomId, q -> q
                .in(BomSpec::getBomId, bomIdList)
                .like(StrUtil.isNotBlank(dto.getColour()), BomSpec::getColour, dto.getColour())
                .like(StrUtil.isNotBlank(dto.getBomSpecCode()), BomSpec::getCode, dto.getBomSpecCode())
                .like(StrUtil.isNotBlank(dto.getBomSpecName()), BomSpec::getName, dto.getBomSpecName()));

        for (BomVo record : records) {
            record.setBomSpecList(bomSpecIdMap.getOrDefault(record.getId(), Collections.emptyList()));
        }

        return page;
    }

    @Override
    public BomVo detail(Long id) {

        BomVo vo = Sql.create(BomVo.class)
                .selectAll(Bom.class)
                .from(Bom.class)
                .eq(Bom::getId, id)
                .one();

        Assert.notNull(vo, "未知bom");

        List<BomSpec> list = bomSpecService.list(q -> q.eq(BomSpec::getBomId, id));
        vo.setBomSpecList(list);

        return vo;
    }

    @DSTransactional
    @Override
    public void add(BomDto dto) {

        dto.setId(IdWorker.getId());

        List<BomSpec> bomSpecList = dto.getBomSpecList().stream()
                .peek(item -> item.setBomId(dto.getId()))
                .map(item -> (BomSpec) item)
                .collect(Collectors.toList());

        // 新增bom
        save(dto);

        // 新增bom明细
        bomSpecService.saveBatch(bomSpecList);
    }

    @DSTransactional
    @Override
    public void edit(BomDto dto) {

        List<BomSpec> bomSpecList = dto.getBomSpecList().stream()
                .peek(item -> item.setBomId(dto.getId()))
                .map(item -> (BomSpec) item)
                .collect(Collectors.toList());

        // 修改bom
        updateById(dto);

        // 修改bom明细
        bomSpecService.editLinked(bomSpecList, BomSpec::getBomId, dto.getId());
    }

    @DSTransactional
    @Override
    public void delete(Long id) {
        removeById(id);
        bomSpecService.remove(q -> q.eq(BomSpec::getBomId, id));
    }

}
