package com.sd.business.service.inventory.impl;

import cn.hutool.core.date.DateUtil;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import com.ruoyi.common.exception.ServiceException;
import com.sd.business.entity.bom.po.Bom;
import com.sd.business.entity.bom.po.BomSpec;
import com.sd.business.entity.department.po.Department;
import com.sd.business.entity.in.dto.InOutStorageDto;
import com.sd.business.entity.in.po.InOutStorageBom;
import com.sd.business.entity.inventory.bo.InOutFun;
import com.sd.business.entity.inventory.dto.InventorySelectDto;
import com.sd.business.entity.inventory.dto.QuantityByDepartmentDto;
import com.sd.business.entity.inventory.dto.QuantityByWarehouseDto;
import com.sd.business.entity.inventory.dto.QuantityDto;
import com.sd.business.entity.inventory.po.Inventory;
import com.sd.business.entity.inventory.vo.InventoryVo;
import com.sd.business.entity.inventory.vo.QuantityByDepartmentVo;
import com.sd.business.entity.inventory.vo.QuantityByWarehouseVo;
import com.sd.business.entity.warehouse.po.Warehouse;
import com.sd.business.mapper.inventory.InventoryMapper;
import com.sd.business.service.bom.BomSpecService;
import com.sd.business.service.inventory.InventoryService;
import com.sd.framework.util.excel.util.ExcelUtil;
import com.sd.framework.util.sql.Sql;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import javax.servlet.http.HttpServletResponse;
import java.math.BigDecimal;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;


/**
 * <p>
 * 库存 服务实现类
 * </p>
 *
 * @author
 * @since 2023-12-11
 */
@Service
public class InventoryServiceImpl extends ServiceImpl<InventoryMapper, Inventory> implements InventoryService {

    @Autowired
    private HttpServletResponse response;

    @Autowired
    private BomSpecService bomSpecService;

    @Override
    public Page<InventoryVo> getPage(InventorySelectDto dto) {

        return Sql.create(InventoryVo.class)
                .selectAll(Inventory.class)
                .selectAs(Warehouse::getName, InventoryVo::getWarehouseName)
                .selectAs(BomSpec::getCode, InventoryVo::getBomSpecCode)
                .selectAs(BomSpec::getName, InventoryVo::getBomSpecName)
                .selectAs(Department::getName, InventoryVo::getDepartmentName)

                .from(Inventory.class)
                .leftJoin(BomSpec.class, Inventory::getBomSpecId)
                .leftJoin(Bom.class, BomSpec::getBomId)
                .leftJoin(Warehouse.class, Inventory::getWarehouseId)
                .leftJoin(Department.class, Inventory::getDepartmentId)

                .eq(Inventory::getDepartmentId, dto.getDepartmentId())
                .eq(Inventory::getWarehouseId, dto.getWarehouseId())
                .in(Inventory::getWarehouseId, dto.getWarehouseIdList())
                .like(BomSpec::getName, dto.getBomSpecName())
                .like(BomSpec::getCode, dto.getBomSpecCode())
                .like(BomSpec::getColour, dto.getBomSpecColour())
                .like(BomSpec::getLength, dto.getBomSpecLength())
                .like(BomSpec::getWidth, dto.getBomSpecWidth())
                .eq(Bom::getBomClassifyId, dto.getBomClassifyId())
                .eq(Bom::getSpecies, dto.getBomSpecies())
                .orderByDesc(Inventory::getId)
                .page(dto);
    }

    @Override
    public void exportExcel(InventorySelectDto dto) {
        dto.setSearchAll(true);
        Page<InventoryVo> page = getPage(dto);
        List<InventoryVo> list = page.getRecords();
        ExcelUtil.export(response, DateUtil.format(new Date(), "yyyy-MM-dd库存"), "库存", list, InventoryVo.class);
    }

    @Override
    public Map<Long, BigDecimal> getQuantity(QuantityDto dto) {
        return mapKV(Inventory::getBomSpecId, Inventory::getQuantity, q -> q
                .eq(Inventory::getWarehouseId, dto.getWarehouseId())
                .eq(Inventory::getDepartmentId, dto.getDepartmentId())
                .in(Inventory::getBomSpecId, dto.getBomSpecIdList()));
    }

    @Override
    public List<QuantityByWarehouseVo> getQuantityByWarehouse(QuantityByWarehouseDto dto) {

        return Sql.create(QuantityByWarehouseVo.class)
                .sum(Inventory::getQuantity, QuantityByWarehouseVo::getInventoryQuantity)
                .selectAs(Warehouse::getId, QuantityByWarehouseVo::getWarehouseId)
                .selectAs(Warehouse::getName, QuantityByWarehouseVo::getWarehouseName)

                .from(Inventory.class)
                .innerJoin(Warehouse.class, Warehouse::getId, Inventory::getId)

                .eq(Inventory::getDepartmentId, dto.getDepartmentId())
                .groupBy(Warehouse::getId)
                .orderByAsc(Warehouse::getId)
                .list();
    }

    @Override
    public List<QuantityByDepartmentVo> getQuantityByDepartment(QuantityByDepartmentDto dto) {

        return Sql.create(QuantityByDepartmentVo.class)
                .sum(Inventory::getQuantity, QuantityByDepartmentVo::getInventoryQuantity)
                .selectAs(Department::getId, QuantityByDepartmentVo::getDepartmentId)
                .selectAs(Department::getName, QuantityByDepartmentVo::getDepartmentName)

                .from(Inventory.class)
                .innerJoin(Department.class, Department::getId, Inventory::getDepartmentId)

                .eq(Inventory::getWarehouseId, dto.getWarehouseId())
                .groupBy(Department::getId)
                .orderByAsc(Department::getId)
                .list();
    }

    @Override
    public void in(InOutStorageDto dto) {
        List<InOutStorageBom> list = dto.getInOutStorageBomList();
        Long departmentId = dto.getDepartmentId();
        Long warehouseId = dto.getWarehouseId();

        List<Long> bomSpecIdList = list.stream().map(InOutFun::getBomSpecId).collect(Collectors.toList());

        synchronized (this) {
            // 通过事业部id和出库id获取bom规格库存
            List<Inventory> inventoryList = getInventoryList(departmentId, warehouseId, bomSpecIdList);

            // 通过bom规格分组
            Map<Long, Inventory> map = inventoryList.stream().collect(Collectors.toMap(Inventory::getBomSpecId, Function.identity()));

            for (InOutFun inOutFun : list) {
                // 库存
                Inventory inventory = map.computeIfAbsent(inOutFun.getBomSpecId(), item -> {
                    Inventory tempInventory = new Inventory();
                    tempInventory.setWarehouseId(warehouseId);
                    tempInventory.setDepartmentId(departmentId);
                    tempInventory.setBomSpecId(item);
                    tempInventory.setQuantity(BigDecimal.ZERO);
                    return tempInventory;
                });
                inventory.setQuantity(inventory.getQuantity().add(inOutFun.getQuantity()));
            }

            saveOrUpdateBatch(map.values());
        }
    }

    @Override
    public void out(InOutStorageDto dto) {
        List<InOutStorageBom> list = dto.getInOutStorageBomList();
        Long departmentId = dto.getDepartmentId();
        Long warehouseId = dto.getWarehouseId();

        List<Long> bomSpecIdList = list.stream().map(InOutFun::getBomSpecId).collect(Collectors.toList());

        synchronized (this) {

            // 通过事业部id和出库id获取bom规格库存
            List<Inventory> inventoryList = getInventoryList(departmentId, warehouseId, bomSpecIdList);
            Map<Long, Inventory> map = inventoryList.stream().collect(Collectors.toMap(Inventory::getBomSpecId, Function.identity()));

            for (InOutFun inOutFun : list) {

                BigDecimal quantity = inOutFun.getQuantity();
                Long bomSpecId = inOutFun.getBomSpecId();
                Inventory inventory = map.get(bomSpecId);

                if (inventory == null || inventory.getQuantity().compareTo(quantity) < 0) {
                    BomSpec bomSpec = bomSpecService.getById(bomSpecId);
                    if (bomSpec == null) {
                        throw new ServiceException("未知bom规格id：" + bomSpecId);
                    }
                    throw new ServiceException("出库失败,品号为：" + bomSpec.getCode() + " 的bom库存不足,库存为：" +
                            (inventory == null ? BigDecimal.ZERO : inventory.getQuantity()) + ",需要出库：" + quantity);
                }
                inventory.setQuantity(inventory.getQuantity().subtract(quantity));
            }

            // 修改库存数量
            updateBatchById(inventoryList);
        }
    }

    /**
     * 通过事业部id和出库id获取bom规格库存
     */
    private List<Inventory> getInventoryList(Long departmentId, Long warehouseId, List<Long> bomSpecIdList) {
        return list(q -> q
                .eq(Inventory::getDepartmentId, departmentId)
                .eq(Inventory::getWarehouseId, warehouseId)
                .in(Inventory::getBomSpecId, bomSpecIdList)
        );
    }

}
