package com.sd.business.util.packing;

import com.sd.business.util.packing.entity.Solution;
import com.sd.business.util.packing.model.Genome;
import com.sd.business.util.packing.model.Item;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Random;

public class GA {

    // 最大的迭代次数
    private static final int maxGen = 10;
    // 变异对换次数
    private static final double variationExchangeCount = 1.00;
    // 复制最优解的次数(选择种群的最优个体，然后复制几次，将最优个体复制多个，存到新的集合中)
    private static final int cloneNumOfBestIndividual = 10;
    // 基因突变概率
    private static final double mutationRate = 0.60;
    // 基因交叉概率
    private static final double crossoverRate = 0.98;
    // 是否可以旋转
    private static final boolean isRotateEnable = true;
    // 随机函数对象
    private static final Random random = new Random();

    // 边界的宽
    private final double width;

    // 边界的长
    private final double length;

    // 矩形数组
    private final Item[] items;

    // 种群数量信息
    private int popSize = 500;

    // 遗传的代数(第几代)
    private int t = 0;

    // 放置所有的种群基因信息
    private List<Genome> population = new ArrayList<>();

    // 新一代的种群基因信息
    private List<Genome> newPopulation = new ArrayList<>();

    // 最好的适应度的对应的染色体(基因序列、路径序列)
    private Genome bestGenome;

    // 各个个体的累积概率
    private double[] probability;

    // 矩形数量
    private int itemNum;

    /**
     * 遗传算法的构造函数
     *
     * @param length 布长
     * @param width  布宽
     * @param items  装箱明细
     */
    private GA(double length, double width, Item[] items) {
        this.length = length;
        this.width = width;
        this.items = items;
    }

    public static Solution solve(double length, double width, Item[] items) {
        GA ga = new GA(length, width, items);
        return ga.solve();
    }

    /**
     * 遗传算法主函数
     *
     * @return 最佳装载结果对象Solution
     */
    private Solution solve() {
        initVar();
        while (t < maxGen) {
            popSize = population.size();
            // 基因进化
            evolution();
            // 更新种群
            population = copyGenomeList(newPopulation);
            t++;
        }
        return bestGenome.getSolution();
    }

    /**
     * 进化函数，正常交叉变异
     */
    private void evolution() {
        //更新累积概率和总适应度
        updateProbabilityAndTotalFitness();
        // 挑选某代种群中适应度最高的个体
        selectBestGenomeAndJoinNext();
        // 赌轮选择策略对剩下的scale-1个个体进行随机的交叉，变异
        while (newPopulation.size() < population.size()) {
            double r = random.nextDouble();
            for (int j = 0; j < probability.length; j++) {
                if (compareDouble(r, probability[j]) != 1) {
                    newPopulation.add(population.get(j));
                    break;
                }
            }
        }
        // 交叉
        if (compareDouble(random.nextDouble(), crossoverRate) != 1) {
            int r = random.nextInt(popSize);
            int k = random.nextInt(popSize);
            while (r == k) {
                r = random.nextInt(popSize);
            }
            cross(k, r);
        }
        // 变异
        for (int i = 0; i < population.size(); i++) {
            if (compareDouble(random.nextDouble(), mutationRate) != 1) {
                variation(i);
            }
        }
    }

    /**
     * 对基因1和基因2进行单点映射交叉
     *
     * @param k1 基因1的索引
     * @param k2 基因2的索引
     */
    private void cross(int k1, int k2) {
        // 获取要交叉的两个基因
        int[] genomeArray1 = newPopulation.get(k1).getGenomeArray();
        int[] genomeArray2 = newPopulation.get(k2).getGenomeArray();
        // 找到交叉位置
        int crossIndex = random.nextInt(itemNum);
        // 获取交叉片段
        for (int i = 0; i <= crossIndex; i++) {
            int temp = genomeArray1[i];
            genomeArray1[i] = genomeArray2[i];
            genomeArray2[i] = temp;
        }
        // 找到重复基因
        HashSet<Integer> set = new HashSet<>();
        // <index,value>
        HashMap<Integer, Integer> repeatMap = new HashMap<>();
        for (int i = 0; i < genomeArray1.length; i++) {
            if (!set.add(genomeArray1[i])) {
                repeatMap.put(i, genomeArray1[i]);
            }
        }
        set.clear();
        for (int i = 0; i < genomeArray2.length; i++) {
            if (!set.add(genomeArray2[i])) {
                for (int key : repeatMap.keySet()) {
                    genomeArray1[key] = genomeArray2[i];
                    genomeArray2[i] = repeatMap.get(key);
                    repeatMap.remove(key);
                    break;
                }
            }
        }
        // 交叉完毕，将基因放回个体，再将个体放回种群，并更新他们的适应值和路径长度
        newPopulation.get(k1).setGenomeArray(genomeArray1);
        newPopulation.get(k1).updateFitnessAndSolution();
        newPopulation.get(k2).setGenomeArray(genomeArray2);
        newPopulation.get(k2).updateFitnessAndSolution();
    }

    /**
     * 变异(n次两两对换)
     *
     * @param k 要进行变异的基因索引
     */
    private void variation(int k) {
        Genome genome = newPopulation.get(k);
        int[] genomeArray = genome.getGenomeArray();
        for (int i = 0; i < variationExchangeCount; i++) {
            int r1 = random.nextInt(itemNum);
            int r2 = random.nextInt(itemNum);
            while (r1 == r2) {
                r2 = random.nextInt(itemNum);
            }
            //交换
            int temp = genomeArray[r1];
            genomeArray[r1] = genomeArray[r2];
            genomeArray[r2] = temp;
        }
        //将变异后的基因序列放回个体
        genome.setGenomeArray(genomeArray);
        // 更新基因的适应值和路程长度
        genome.updateFitnessAndSolution();
        //将变异后的个体放回种群
        newPopulation.set(k, genome);
    }

    /**
     * 初始化函数，进行一些初始化操作
     */
    private void initVar() {
        this.itemNum = this.items.length;
        this.population = new ArrayList<>();
        //初始化种群信息
        List<Integer> sequence = new ArrayList<>();
        for (int i = 0; i < itemNum; i++) {
            sequence.add(i);
        }
        for (int i = 0; i < popSize; i++) {
            Collections.shuffle(sequence);
            int[] initSequence = new int[itemNum];
            for (int j = 0; j < sequence.size(); j++) {
                initSequence[j] = sequence.get(j);
            }
            Genome genome = new Genome(length, width, items, isRotateEnable, initSequence.clone());
            genome.updateFitnessAndSolution();
            population.add(genome);
        }
        bestGenome = copyGenome(population.get(0));
        for (int i = 1; i < popSize; i++) {
            Genome genome = population.get(i);
            if (bestGenome.getFitness() < genome.getFitness()) {
                bestGenome = copyGenome(genome);
            }
        }
        System.out.println("初始解为：" + bestGenome.getSolution().getUseRatio());
    }

    /**
     * 挑选种群中适应度最大的个体基因，并复制n份，直接加入下一代种群
     */
    private void selectBestGenomeAndJoinNext() {
        newPopulation = new ArrayList<>();
        Genome tempBest = population.get(0);
        for (int i = 1; i < population.size(); i++) {
            if (population.get(i).getFitness() > tempBest.getFitness()) {
                tempBest = population.get(i);
            }
        }
        if (compareDouble(tempBest.getFitness(), bestGenome.getFitness()) == 1) {
            bestGenome = copyGenome(tempBest);
            // 最佳迭代次数
            System.out.println("当前代数: " + t + " : " + bestGenome.getSolution().getUseRatio());
        }
        for (int i = 0; i < cloneNumOfBestIndividual; i++) {
            newPopulation.add(copyGenome(tempBest));
        }
    }

    /**
     * 更新各个个体的累积概率和种群总适应度
     */
    private void updateProbabilityAndTotalFitness() {
        probability = new double[population.size()];
        // 种群总的适应度数值
        double totalFitness = 0;
        for (Genome genome : population) {
            totalFitness += genome.getFitness();
        }
        double rate = 0.0;
        for (int i = 0; i < population.size(); i++) {
            rate += (population.get(i).getFitness() / totalFitness);
            probability[i] = rate;
        }
    }

    /**
     * 根据传入的模板基因，进行拷贝，返回拷贝好的基因
     *
     * @param genome 模板基因
     * @return 返回拷贝好的基因
     */
    private Genome copyGenome(Genome genome) {
        Genome copy = new Genome(genome.getLength(), genome.getWidth(), genome.getItems(), genome.isRotateEnable(), genome.getGenomeArray().clone());
        copy.setSolution(genome.getSolution());
        copy.setFitness(genome.getFitness());
        return copy;
    }

    /**
     * 根据传入的模板基因集合，进行拷贝，返回拷贝好的基因集合
     *
     * @param genomeList 模板基因集合
     * @return 返回拷贝好的基因集合
     */
    private List<Genome> copyGenomeList(List<Genome> genomeList) {
        List<Genome> copyList = new ArrayList<>();
        for (Genome genome : genomeList) {
            copyList.add(copyGenome(genome));
        }
        return copyList;
    }

    /**
     * 判断两个双精度浮点型变量的大小关系
     *
     * @param d1 双精度浮点型变量1
     * @param d2 双精度浮点型变量2
     * @return 返回0代表两个数相等，返回1代表前者大于后者，返回-1代表前者小于后者，
     */
    private int compareDouble(double d1, double d2) {
        // 定义一个误差范围，如果两个数相差小于这个误差，则认为他们是相等的 1e-06 = 0.000001
        double error = 1e-06;
        if (Math.abs(d1 - d2) < error) {
            return 0;
        } else if (d1 < d2) {
            return -1;
        } else if (d1 > d2) {
            return 1;
        } else {
            throw new RuntimeException("d1 = " + d1 + " , d2 = " + d2);
        }
    }

}
