package com.sd.framework.util.sql;

import cn.hutool.core.util.ReflectUtil;
import cn.hutool.core.util.StrUtil;
import com.baomidou.mybatisplus.annotation.TableField;
import com.baomidou.mybatisplus.core.toolkit.LambdaUtils;
import com.baomidou.mybatisplus.core.toolkit.StringPool;
import com.baomidou.mybatisplus.core.toolkit.support.SFunction;
import org.apache.ibatis.reflection.property.PropertyNamer;

import java.lang.reflect.Field;
import java.util.Arrays;
import java.util.Set;
import java.util.stream.Collectors;

public class Select<T> {

    private final Sql<T> sql;

    protected Select(Sql<T> sql) {
        this.sql = sql;
    }

    public <S, V> Select<T> select(SFunction<S, V> function) {
        sql.selectList.add(sql.getSqlFieldName(function));
        return this;
    }

    public <S, V> Select<T> select(String alias, SFunction<S, V> function) {
        sql.selectList.add(sql.getSqlFieldName(alias, function));
        return this;
    }

    @SafeVarargs
    public final <S, V> Select<T> select(SFunction<S, V>... functions) {
        for (SFunction<S, V> function : functions) {
            select(function);
        }
        return this;
    }

    @SafeVarargs
    public final <S, V> Select<T> select(String alias, SFunction<S, V>... functions) {
        for (SFunction<S, V> function : functions) {
            select(alias, function);
        }
        return this;
    }

    public <S, V> Select<T> selectAs(SFunction<S, V> function, SFunction<T, ?> asFunction) {
        String as = PropertyNamer.methodToProperty(LambdaUtils.extract(asFunction).getImplMethodName());
        sql.selectList.add(sql.getSqlFieldName(function) + StringPool.SPACE + as);
        return this;
    }

    public <S, V> Select<T> selectAs(String alias, SFunction<S, V> function, SFunction<T, ?> asFunction) {
        String as = PropertyNamer.methodToProperty(LambdaUtils.extract(asFunction).getImplMethodName());
        sql.selectList.add(sql.getSqlFieldName(alias, function) + StringPool.SPACE + as);
        return this;
    }

    public <S, V> Select<T> selectAs(String selectSql, SFunction<T, ?> asFunction) {
        String as = PropertyNamer.methodToProperty(LambdaUtils.extract(asFunction).getImplMethodName());
        sql.selectList.add(selectSql + StringPool.SPACE + as);
        return this;
    }

    @SafeVarargs
    public final <S> Select<T> selectAll(Class<S> cls, SFunction<S, ?>... excludeFunction) {
        return selectAll(sql.getTableAlias(cls), cls, excludeFunction);
    }

    @SafeVarargs
    public final <S> Select<T> selectAll(String alias, Class<S> cls, SFunction<S, ?>... excludeFunction) {
        Set<String> excludeNameSet = Arrays.stream(excludeFunction)
                .map(item -> PropertyNamer.methodToProperty(LambdaUtils.extract(item).getImplMethodName()))
                .collect(Collectors.toSet());

        Field[] fields = ReflectUtil.getFields(cls);
        for (Field field : fields) {
            String fieldName = ReflectUtil.getFieldName(field);
            if ("serialVersionUID".equals(fieldName)) {
                continue;
            }
            if (excludeNameSet.contains(fieldName)) {
                continue;
            }
            TableField tableField = field.getAnnotation(TableField.class);
            if (tableField != null) {
                if (!tableField.exist() || !tableField.select()) {
                    continue;
                }
                String value = tableField.value();
                if (StrUtil.isNotBlank(value)) {
                    fieldName = value;
                }
            }
            sql.selectList.add(alias + StringPool.DOT + StrUtil.toUnderlineCase(fieldName));
        }
        return this;
    }

    public Select<T> distinct() {
        sql.distinct = true;
        return this;
    }

    public <S, V> Select<T> max(SFunction<S, V> function, SFunction<T, ?> asFunction) {
        String as = PropertyNamer.methodToProperty(LambdaUtils.extract(asFunction).getImplMethodName());
        sql.selectList.add("MAX(" + sql.getSqlFieldName(function) + ") " + as);
        return this;
    }

    public <S, V> Select<T> max(String alias, SFunction<S, V> function, SFunction<T, ?> asFunction) {
        String as = PropertyNamer.methodToProperty(LambdaUtils.extract(asFunction).getImplMethodName());
        sql.selectList.add("MAX(" + sql.getSqlFieldName(alias, function) + ") " + as);
        return this;
    }

    public <S, V> Select<T> min(SFunction<S, V> function, SFunction<T, ?> asFunction) {
        String as = PropertyNamer.methodToProperty(LambdaUtils.extract(asFunction).getImplMethodName());
        sql.selectList.add("MIN(" + sql.getSqlFieldName(function) + ") " + as);
        return this;
    }

    public <S, V> Select<T> min(String alias, SFunction<S, V> function, SFunction<T, ?> asFunction) {
        String as = PropertyNamer.methodToProperty(LambdaUtils.extract(asFunction).getImplMethodName());
        sql.selectList.add("MIN(" + sql.getSqlFieldName(alias, function) + ") " + as);
        return this;
    }

    public <S, V> Select<T> count(SFunction<S, V> function, SFunction<T, ?> asFunction) {
        String as = PropertyNamer.methodToProperty(LambdaUtils.extract(asFunction).getImplMethodName());
        sql.selectList.add("COUNT(" + sql.getSqlFieldName(function) + ") " + as);
        return this;
    }

    public <S, V> Select<T> count(String alias, SFunction<S, V> function, SFunction<T, ?> asFunction) {
        String as = PropertyNamer.methodToProperty(LambdaUtils.extract(asFunction).getImplMethodName());
        sql.selectList.add("COUNT(" + sql.getSqlFieldName(alias, function) + ") " + as);
        return this;
    }

    public <S, V> Select<T> sum(SFunction<S, V> function, SFunction<T, ?> asFunction) {
        String as = PropertyNamer.methodToProperty(LambdaUtils.extract(asFunction).getImplMethodName());
        sql.selectList.add("SUM(" + sql.getSqlFieldName(function) + ") " + as);
        return this;
    }

    public <S, V> Select<T> sum(String alias, SFunction<S, V> function, SFunction<T, ?> asFunction) {
        String as = PropertyNamer.methodToProperty(LambdaUtils.extract(asFunction).getImplMethodName());
        sql.selectList.add("SUM(" + sql.getSqlFieldName(alias, function) + ") " + as);
        return this;
    }

    public <S, V> Select<T> add(SFunction<S, V> function1, SFunction<S, V> function2, SFunction<T, ?> asFunction) {
        String as = PropertyNamer.methodToProperty(LambdaUtils.extract(asFunction).getImplMethodName());
        sql.selectList.add(sql.getSqlFieldName(function1) + "+" + sql.getSqlFieldName(function2) + " " + as);
        return this;
    }

    public <S, V> Select<T> subtract(SFunction<S, V> function1, SFunction<S, V> function2, SFunction<T, ?> asFunction) {
        String as = PropertyNamer.methodToProperty(LambdaUtils.extract(asFunction).getImplMethodName());
        sql.selectList.add(sql.getSqlFieldName(function1) + "-" + sql.getSqlFieldName(function2) + " " + as);
        return this;
    }

    public <S, V> Select<T> multiply(SFunction<S, V> function1, SFunction<S, V> function2, SFunction<T, ?> asFunction) {
        String as = PropertyNamer.methodToProperty(LambdaUtils.extract(asFunction).getImplMethodName());
        sql.selectList.add(sql.getSqlFieldName(function1) + "*" + sql.getSqlFieldName(function2) + " " + as);
        return this;
    }

    public <S, V> Select<T> divide(SFunction<S, V> function1, SFunction<S, V> function2, SFunction<T, ?> asFunction) {
        String as = PropertyNamer.methodToProperty(LambdaUtils.extract(asFunction).getImplMethodName());
        sql.selectList.add(sql.getSqlFieldName(function1) + "/" + sql.getSqlFieldName(function2) + " " + as);
        return this;
    }

    public From<T> from(Class<?> cls) {
        String tableAlias = sql.getTableAlias(cls);
        return from(tableAlias, cls);
    }

    public From<T> from(String alias, Class<?> cls) {
        String tableName = Sql.getTableName(cls);
        sql.fromList.add(" FROM " + tableName + StringPool.SPACE + alias);
        return new From<>(sql);
    }

}
